package com.contentsquare.flutter

import android.app.Activity
import android.content.Context
import android.view.View
import android.view.ViewGroup
import androidx.annotation.NonNull
import androidx.lifecycle.DefaultLifecycleObserver
import androidx.lifecycle.LifecycleOwner
import com.contentsquare.android.api.bridge.flutter.ExternalSessionReplayListener
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import com.contentsquare.android.api.bridge.flutter.ExternalViewGraphListener
import com.contentsquare.android.api.bridge.flutter.ExternalViewGraphResult
import com.contentsquare.android.api.bridge.flutter.FlutterInterface
import com.contentsquare.android.api.bridge.xpf.ExternalBridgeInterface
import com.contentsquare.flutter.infrastructure.bridges.FlutterInterfaceBridge
import com.contentsquare.flutter.infrastructure.core.method_call_executor.MethodCallExecutor
import com.contentsquare.flutter.infrastructure.core.method_call_executor.agents.MethodCallAgent
import com.contentsquare.flutter.infrastructure.core.method_call_executor.data.ExecutionData
import io.flutter.Log
import io.flutter.embedding.android.FlutterImageView
import io.flutter.embedding.android.FlutterView
import io.flutter.embedding.engine.mutatorsstack.FlutterMutatorView
import io.flutter.embedding.engine.plugins.lifecycle.HiddenLifecycleReference

internal class FlutterViewFinder {

    fun find(activity: Activity): View? {
        val view = activity.window.decorView
        return findRecursively(view)
    }

    private fun findRecursively(view: View): View? {
        if (view is FlutterView) {
            return view
        }
        if (view is ViewGroup) {
            for (childIndex in 0 until view.childCount) {
                val viewFound = findRecursively(view.getChildAt(childIndex))
                if (viewFound != null) {
                    return viewFound
                }
            }
        }
        return null
    }
}

private const val REQUEST_ENABLE_API_ERROR_METHOD = "requestEnableAPIError"
private const val REQUEST_ENABLE_SR_METHOD = "requestEnableSessionReplay"
private const val REQUEST_INITIAL_CONFIG = "requestInitialConfig"
private const val ENABLE_SR_KEY = "enableSR"
private const val ENABLE_API_ERRORS_KEY = "enableApiErrors"

/** ContentsquarePlugin */
class ContentsquarePlugin : FlutterPlugin, ActivityAware, MethodCallHandler,
    DefaultLifecycleObserver {
    private val flutterTVPSeparator = "|flutter|"

    private lateinit var channel: MethodChannel
    private var context: Context? = null
    private lateinit var view: View
    private lateinit var activity: Activity
    private lateinit var executor: MethodCallExecutor
    private var flutterInterfaceBridge = FlutterInterfaceBridge.default()
    private val externalBridgeInterface = object : ExternalBridgeInterface {
        override fun enableApiErrors(enable: Boolean) {
            channel.invokeMethod(
                REQUEST_ENABLE_API_ERROR_METHOD,
                mapOf(ENABLE_API_ERRORS_KEY to enable)
            )
        }

        override fun enableSessionReplay(enable: Boolean) {
            channel.invokeMethod(REQUEST_ENABLE_SR_METHOD, mapOf(ENABLE_SR_KEY to enable))
        }

        override fun notifyAboutInitialConfig(apiErrorsEnabled: Boolean, srEnabled: Boolean) {
            channel.invokeMethod(
                REQUEST_INITIAL_CONFIG,
                mapOf(
                    ENABLE_SR_KEY to srEnabled,
                    ENABLE_API_ERRORS_KEY to apiErrorsEnabled,
                ),
            )
        }
    }

    private fun registerNewExternalViewFromActivity(@NonNull activity: Activity) {
        view = activity.findViewById(android.R.id.content) as View


        val flutterView = FlutterViewFinder().find(activity)
        if (flutterView == null) {
            Log.d("Android", "ContentsquarePlugin, no Flutter view found.")
            return
        }
        Log.d("Android", "ContentsquarePlugin, Flutter view found $flutterView")

        FlutterInterface.excludeExternalView(
            FlutterImageView::class.java,
            FlutterMutatorView::class.java,
        );

        // We register the [FlutterView] as the Flutter root's view even if it is not be the most
        // deepest view on Android side before getting to Flutter part.
        // The reason is that Flutter adds, below the [FlutterView], either a [FlutterSurfaceView]
        // or a [FlutterTextureView] depending on the BackgroundMode of the [FlutterActivity].
        // In order to be sure that we handle both cases (and also potentials future cases of other
        // deeper views), we consider the [FlutterView] as the root of the content created by Flutter.
        flutterInterfaceBridge.registerExternalBridge(externalBridgeInterface)
        flutterInterfaceBridge.registerExternalView(flutterView,
            object : ExternalViewGraphListener {
                override fun takeSnapShot(
                    view: View, path: String, result: ExternalViewGraphResult
                ) {
                    val externalViewGraphResult = result

                    val targetViewPath: String = path + flutterTVPSeparator

                    val pixelRatio: Double =
                        context?.resources?.displayMetrics?.density?.toDouble() ?: 1.0

                    channel.invokeMethod("requestFlutterSnapshot",
                        mapOf("rootPath" to targetViewPath, "pixelRatio" to pixelRatio),
                        object : Result {
                            override fun success(result: Any?) {
                                externalViewGraphResult.onSnapshotTaken(
                                    view, result as String
                                )
                            }

                            override fun error(
                                errorCode: String, errorMessage: String?, errorDetails: Any?
                            ) {
                                Log.d("Android", "$errorCode, $errorMessage, $errorDetails")
                            }

                            override fun notImplemented() {
                                Log.d("Android", "notImplemented")
                            }
                        })
                }
            })
    }

    override fun onAttachedToEngine(
        @NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding
    ) {
        channel = MethodChannel(flutterPluginBinding.binaryMessenger, "contentsquare")
        channel.setMethodCallHandler(this)
        context = flutterPluginBinding.applicationContext

        executor = MethodCallExecutor()
    }

    override fun onAttachedToActivity(@NonNull activityPluginBinding: ActivityPluginBinding) {
        (activityPluginBinding.lifecycle as HiddenLifecycleReference).lifecycle.addObserver(this)
        activity = activityPluginBinding.activity
    }

    override fun onResume(owner: LifecycleOwner) {
        registerNewExternalViewFromActivity(activity)
    }

    override fun onReattachedToActivityForConfigChanges(@NonNull activityPluginBinding: ActivityPluginBinding) {
        registerNewExternalViewFromActivity(activityPluginBinding.activity)
    }

    override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {
        val agent = MethodCallAgent.of(call.method)
        val data = ExecutionData(call = call, context = context)
        executor.execute(agent = agent, data = data, result = result)
    }

    override fun onDetachedFromEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        channel.setMethodCallHandler(null)
        context = null
    }

    override fun onDetachedFromActivityForConfigChanges() {
        flutterInterfaceBridge.unRegisterExternalView(view)
        flutterInterfaceBridge.unregisterExternalBridge(externalBridgeInterface)
    }

    override fun onDetachedFromActivity() {
        flutterInterfaceBridge.unRegisterExternalView(view)
        flutterInterfaceBridge.unregisterExternalBridge(externalBridgeInterface)
    }
}
