package com.contentsquare.flutter.infrastructure.bridges

import android.view.View
import com.contentsquare.android.api.bridge.flutter.ExternalSessionReplayListener
import com.contentsquare.android.api.bridge.flutter.ExternalViewGraphListener
import com.contentsquare.android.api.bridge.flutter.FlutterInterface
import com.contentsquare.android.api.bridge.xpf.ExternalBridgeInterface
import com.contentsquare.android.api.bridge.xpf.XpfInterface

/**
 * Bridge to FlutterInterface of contentsquare.
 */
internal interface FlutterInterfaceBridge {

    /**
     * Send event.
     *
     * @param json the json of the event.
     */
    fun sendEvent(json: String)

    /**
     * Register a view.
     *
     * @param view the view to register
     * @param listener the listener of some events
     */
    fun registerExternalView(view: View, listener: ExternalViewGraphListener)

    /**
     * Unregister a view.
     *
     * @param view the view to unregister
     */
    fun unRegisterExternalView(view: View)

    /**
     * Register a flutter session replay listener to inform
     * when the session replay is started or stopped.
     *
     * @param listener the listener of session replay state
     */
    fun registerFlutterSessionReplayListener(listener: ExternalSessionReplayListener)

    /**
     * Unregister a flutter session replay listener.
     *
     * @param listener the listener of session replay state
     */
    fun unregisterFlutterSessionReplayListener(listener: ExternalSessionReplayListener)

    /**
     * Send session replay event.
     *
     * @param event the event to send
     * @param pixelRatio the pixel ratio of the device
     */
    fun sendSessionReplayEvent(event: String, pixelRatio: Float)

    /**
     * Register an external bridge for the communication with the native SDK.
     *
     * @param externalBridge local implementation of the [ExternalBridgeInterface].
     */
    fun registerExternalBridge(externalBridge: ExternalBridgeInterface)

    /**
     * Unregister an external bridge for the communication with the native SDK.
     *
     * @param externalBridge local implementation of the [ExternalBridgeInterface].
     */
    fun unregisterExternalBridge(externalBridge: ExternalBridgeInterface)

    companion object Factory {
        /**
         * Creates a new instance of the default implementation of [FlutterInterfaceBridge].
         */
        fun default(): FlutterInterfaceBridge {
            return FlutterInterfaceBridgeImpl()
        }
    }
}


/**
 * Channel of method in the contentsquare sdk flutter interface.
 */
internal enum class FlutterInterfaceChannel(val value: String) {
    /** channel/method called to send event */
    SEND_EVENT("sendEvent"),

    /** channel/method called to send session replay event */
    SEND_SESSION_REPLAY_EVENT("sendSessionReplayEvent"),

    /** channel/method called to register view */
    EXTERNAL_VIEW_REGISTERER("registerExternalView"),

    /** channel/method called to unregister view */
    EXTERNAL_VIEW_UNREGISTERER("unRegisterExternalView"),

    /** channel/method called to register flutter sr listener */
    SESSION_REPLAY_LISTENER_REGISTERER("registerFlutterSrListener"),

    /** channel/method called to unregister flutter sr listener */
    SESSION_REPLAY_LISTENER_UNREGISTERER("unregisterFlutterSrListener"),

    /**
     * Channel/method called to register external bridge.
     *
     * See [XpfInterface.registerExternalBridge].
     */
    REGISTER_EXTERNAL_BRIDGE("registerExternalBridge"),

    /**
     * Channel/method called to unregister external bridge.
     *
     * See [XpfInterface.unregisterExternalBridge].
     */
    UNREGISTER_EXTERNAL_BRIDGE("unregisterExternalBridge")
}


/**
 * Implementation of [FlutterInterfaceBridge].
 *
 * @see FlutterInterfaceBridge
 */
internal class FlutterInterfaceBridgeImpl : FlutterInterfaceBridge {

    private val objectInstanceName = "INSTANCE"

    override fun sendEvent(json: String) {
        val flutterInterface = FlutterInterface::class.java
        flutterInterface.getDeclaredMethod(
            FlutterInterfaceChannel.SEND_EVENT.value,
            String::class.java
        ).run {
            isAccessible = true
            invoke(null, json)
        }
    }

    override fun registerExternalView(view: View, listener: ExternalViewGraphListener) {
        val flutterInterface = FlutterInterface::class.java
        flutterInterface.getDeclaredMethod(
            FlutterInterfaceChannel.EXTERNAL_VIEW_REGISTERER.value,
            View::class.java,
            ExternalViewGraphListener::class.java
        ).run {
            isAccessible = true
            invoke(null, view, listener)
        }
    }

    override fun unRegisterExternalView(view: View) {
        val flutterInterface = FlutterInterface::class.java
        flutterInterface.getDeclaredMethod(
            FlutterInterfaceChannel.EXTERNAL_VIEW_UNREGISTERER.value, View::class.java
        ).run {
            isAccessible = true
            invoke(null, view)
        }
    }

    override fun registerFlutterSessionReplayListener(listener: ExternalSessionReplayListener) {
            val flutterInterface = FlutterInterface::class.java
            flutterInterface.getDeclaredMethod(
                    FlutterInterfaceChannel.SESSION_REPLAY_LISTENER_REGISTERER.value, ExternalSessionReplayListener::class.java
            ).run {
                isAccessible = true
                invoke(null, listener)
            }
        }

    override fun unregisterFlutterSessionReplayListener(listener: ExternalSessionReplayListener) {
            val flutterInterface = FlutterInterface::class.java
            flutterInterface.getDeclaredMethod(
                    FlutterInterfaceChannel.SESSION_REPLAY_LISTENER_UNREGISTERER.value, ExternalSessionReplayListener::class.java
            ).run {
                isAccessible = true
                invoke(null, listener)
            }
    }

    override fun sendSessionReplayEvent(event: String, pixelRatio: Float) {
        val flutterInterface = FlutterInterface::class.java
        flutterInterface.getDeclaredMethod(
                FlutterInterfaceChannel.SEND_SESSION_REPLAY_EVENT.value, String::class.java, Float::class.java
        ).run {
            isAccessible = true
            invoke(null, event, pixelRatio)
        }
    }
    override fun registerExternalBridge(externalBridge: ExternalBridgeInterface) {
        val xpfInterface = XpfInterface::class.java
        xpfInterface.getDeclaredMethod(
            FlutterInterfaceChannel.REGISTER_EXTERNAL_BRIDGE.value,
            ExternalBridgeInterface::class.java
        ).run {
            isAccessible = true
            val instance = xpfInterface.getDeclaredField(objectInstanceName).get(null)
            invoke(instance, externalBridge)
        }
    }

    override fun unregisterExternalBridge(externalBridge: ExternalBridgeInterface) {
        val xpfInterface = XpfInterface::class.java
        xpfInterface.getDeclaredMethod(
            FlutterInterfaceChannel.UNREGISTER_EXTERNAL_BRIDGE.value,
            ExternalBridgeInterface::class.java
        ).run {
            isAccessible = true
            val instance = xpfInterface.getDeclaredField(objectInstanceName).get(null)
            invoke(instance, externalBridge)
        }
    }

}
