/// Agent used to execute specific method from `FlutterMethodCall`.
protocol MethodCallAgent {
    /// Message for the error not caught internaly.
    var nativeExceptionMessage: String? { get }
    
    /// Process the execution of the method.
    ///
    /// - Parameter data: Data used during execution
    func execute(data: ExecutionData) throws -> ExecutionResponse
}

/// Error thrown during the execution.
struct FailedExecutionError: Error {
    /// `FailedExecutionResponseReason`  of the error
    let reason: FailedExecutionResponseReason
    
    /// Message explaining the error
    let message: String?
    
    /// Cause of the error
    let cause: Any?
}

/// Extension on `MethodCallAgent` to add execution in safe and catch errors.
extension MethodCallAgent {
    
    ///Process the execution of the method and catch any error/exception.
    func safeExecute(data: ExecutionData) -> ExecutionResponse {
        do {
            return try execute(data: data)
        } catch let error as FailedExecutionError {
            return ExecutionResponse.failed(reason: error.reason, message: error.message, cause: error.cause)
        } catch {
            return ExecutionResponse.failed(reason: FailedExecutionResponseReason.nativeException, message: nativeExceptionMessage, cause: error)
        }
    }
}

/// Enumerations of all handled method calls
enum MethodCallAgentMethod: String, CaseIterable {
    case unknownMethod = "__unknown_method__"
    case collectCustomTelemetryData = "collectCustomTelemetryData"
    case collectFlutterVersion = "collectFlutterVersion"
    case forgetMe = "forgetMe"
    case getCurrentSessionReplayUrl = "currentSessionReplayUrl"
    case getUserId = "getUserId"
    case optIn = "optIn"
    case optOut = "optOut"
    case resumeTracking = "resumeTracking"
    case sendApiError = "sendApiError"
    case sendDynamicVar = "sendDynamicVar"
    case sendScreenName = "send"
    case sendSessionReplayEvent = "sendSessionReplayEvent"
    case sendTouchEvent = "sendTouchEvent"
    case sendTransaction = "sendTransaction"
    case setURLMaskingPatterns = "setURLMaskingPatterns"
    case stopTracking = "stopTracking"
}

/// Factory creator class for [MethodCallAgent].
struct MethodCallAgentFactory {
    
    /// Creates a new instance of [MethodCallAgent].
    ///
    /// - Parameter method: The method name to use to create the good `MethodCallAgent`
    static func makeInstance(method: String) -> MethodCallAgent {
        var agentMethod = MethodCallAgentMethod.allCases.first(where: { (agentMethod) -> Bool in
            return agentMethod.rawValue == method
        })
        
        if (agentMethod == nil) {
            agentMethod = MethodCallAgentMethod.unknownMethod
        }
        
        switch (agentMethod) {
        case .collectCustomTelemetryData:
            return CollectCustomTelemetryAgent()
        case .collectFlutterVersion:
            return CollectFlutterVersionAgent()
        case .forgetMe:
            return ForgetMeAgent()
        case .getCurrentSessionReplayUrl:
            return GetCurrentSessionReplayLinkAgent()
        case .getUserId:
            return GetUserIdAgent()
        case .optIn:
            return OptInAgent()
        case .optOut:
            return OptOutAgent()
        case .resumeTracking:
            return ResumeTrackingAgent()
        case .sendApiError:
            return SendApiErrorAgent()
        case .sendDynamicVar:
            return SendDynamicVarAgent()
        case .sendScreenName:
            return SendScreenNameAgent()
        case .sendSessionReplayEvent:
            return SendSessionReplayEventAgent()
        case .sendTouchEvent:
            return SendTouchEventAgent()
        case .sendTransaction:
            return SendTransactionAgent()
        case .setURLMaskingPatterns:
            return SetURLMaskingPatternsAgent()
        case .stopTracking:
            return StopTrackingAgent()
        case .unknownMethod:
            return MissingMethodCallAgent(method: method)
        case .none:
            return MissingMethodCallAgent(method: method)
        }
    }
}
