import Flutter
import UIKit
import ContentsquareModule
import ContentsquareErrorAnalysisModule

// This is the protocol that the listeners passed by external bridges have to implement.
@objc
protocol ExternalBridgeInterface {
    @objc
    func takeSnapshot(parameters: [String: Any])
    
    @objc
    func enableSessionReplay(parameters: [String: Any])

    @objc
    func enableAPIErrors(parameters: [String: Any])
}

private typealias TakeSnapshotHandler = (_ rootPath: String, _ completion: @escaping (String?) -> Void) -> Void
private typealias EnableSessionReplayHandler = (_ enable: Bool) -> Void
private typealias EnableAPIErrorHandler = (_ enable: Bool) -> Void

private class ExternalBridgeImpl: ExternalBridgeInterface {
    private var _takeSnapshotHandler: TakeSnapshotHandler
    private var _enableSessionReplayHandler: EnableSessionReplayHandler
    private var _enableAPIErrorHandler: EnableAPIErrorHandler
    
    init(takeSnapshotHandler: @escaping TakeSnapshotHandler, enableSessionReplayHandler: @escaping EnableSessionReplayHandler, enableAPIErrorHandler: @escaping EnableAPIErrorHandler) {
        _takeSnapshotHandler = takeSnapshotHandler
        _enableSessionReplayHandler = enableSessionReplayHandler
        _enableAPIErrorHandler = enableAPIErrorHandler
    }
    
    @objc
    func takeSnapshot(parameters: [String: Any]) {
        if let rootPath = parameters[TakeSnapshotParam.rootPath.rawValue] as? String,
           let completion = parameters[TakeSnapshotParam.completion.rawValue] as? (String?) -> Void {
            _takeSnapshotHandler(rootPath, completion)
        }
    }
    
    @objc
    func enableSessionReplay(parameters: [String: Any]) {
        if let enable = parameters[EnableSessionReplayParam.enable.rawValue] as? Bool {
            _enableSessionReplayHandler(enable)
        }
    }

    func enableAPIErrors(parameters: [String : Any]) {
        if let enable = parameters[EnableAPIErrorParam.enable.rawValue] as? Bool {
            _enableAPIErrorHandler(enable)
        }
    }
}

// Expected elements takeSnapshot's parameters dictionary:
enum TakeSnapshotParam: String {
    
    // path of the rootView of the bridge on the native side. To be used to create TargetViewPaths when generating the snapshot
    case rootPath // as String - mandatory
    
    // completion block to be used in order to return the json of the generated snapshot asynchronoulsy
    case completion // as (json:String?) -> Void - mandatory
}

enum EnableSessionReplayParam: String {
    
    // flag, that represents the new status of the session recording.
    case enable // as bool - mandatory
}

enum EnableAPIErrorParam: String {

    // flag, that represents the new status of the session recording.
    case enable // as bool - mandatory
}

enum FlutterChannelMethod: String {
    case takeSnapshot = "requestFlutterSnapshot"
    case enableSR = "requestEnableSessionReplay"
    case enableAPIError = "requestEnableAPIError"
}

func registerExternalBridge(channel: FlutterMethodChannel) {
    let flutterTag = "|flutter|"
    
    func snapshotHandler(rootPath: String, completion: @escaping (String?) -> Void) {
        let arguments: [String: Any] = [TakeSnapshotParam.rootPath.rawValue: rootPath + flutterTag]
        channel.invokeMethod(FlutterChannelMethod.takeSnapshot.rawValue, arguments: arguments) { result in
            var json = result as? String
            completion(json)
        }
    }
    
    func enableSRHandler(enable: Bool) {
        let arguments: [String: Any] = ["enableSR": enable]
        channel.invokeMethod(FlutterChannelMethod.enableSR.rawValue, arguments: arguments)
    }

    func enableAPIErrorHandler(enable: Bool) {
        let arguments: [String: Any] = ["enableApiErrors": enable]
        channel.invokeMethod(FlutterChannelMethod.enableAPIError.rawValue, arguments: arguments)
     }
    
    let interface = ExternalBridgeImpl(
        takeSnapshotHandler: snapshotHandler,
        enableSessionReplayHandler: enableSRHandler,
        enableAPIErrorHandler: enableAPIErrorHandler
    )
    
    Contentsquare.perform(
        NSSelectorFromString("_registerExternalBridgeWithParameters:"),
        with: [
            "interface": interface,
            "type": 1
        ]
    )
}

public class SwiftContentsquarePlugin: NSObject, FlutterPlugin {
    private var executor: MethodCallExecutor
    static private var _channel: FlutterMethodChannel!
    
    init(executor: MethodCallExecutor = MethodCallExecutor()) {
        self.executor = executor
    }
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        if (SwiftContentsquarePlugin._channel != nil) {
            NSLog(
            """
            CSLIB ⚠️ WARNING: SwiftContentsquarePlugin was registered twice. Please refer to the Contentsquare documentation's troubleshooting section for more information about what to do: https://docs.contentsquare.com/flutter/#troubleshooting
            """
            );
            return;
        }
        let channel = FlutterMethodChannel(name: "contentsquare", binaryMessenger: registrar.messenger())
        SwiftContentsquarePlugin._channel = channel
        
        let instance = SwiftContentsquarePlugin()
        registrar.addMethodCallDelegate(instance, channel: channel)
        registerExternalBridge(channel: channel)
    }
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        let agent: MethodCallAgent = MethodCallAgentFactory.makeInstance(method: call.method)
        let executionData = ExecutionData(call: call)
        
        self.executor.execute(agent: agent, data: executionData, result: result)
    }
}

extension UIView {
    
    func allSubviews(where predicate: (UIView) -> Bool) -> [UIView] {
        let views: [UIView] = predicate(self) ? [self] : []
        return views + subviews.flatMap { $0.allSubviews(where: predicate) }
    }
    
    var className: String {
        String(describing: type(of: self))
    }
}
