import 'package:contentsquare/src/hardball1921.dart';
import 'package:equatable/equatable.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';

/// A class which determines what should be masked during Session Replay.
///
/// See also:
///   * [SessionReplayMaskingScope] for use of [MaskingConfig]
class MaskingConfig extends Equatable {
  /// If true, any widget that creates a [RenderParagraph] will be masked
  /// during Session Replay. E.g: [Text], [RichText]
  final bool? maskTexts;

  /// If true, any widget that creates a [RenderEditable] will be masked
  /// during Session Replay. E.g: [TextFormField], [TextField]
  final bool? maskTextFields;

  /// If true, any widget that creates a [RenderImage] will be masked
  /// during Session Replay. E.g: [Image]
  final bool? maskImages;

  /// A factory used to create a custom [MaskingConfig]
  ///
  /// For easier readability, consider using:
  ///  * [MaskingConfig.maskAll] -> If setting all args to true.
  ///  * [MaskingConfig.unMaskAll] -> If setting all args to false.
  ///
  /// This factory does not accept explicitly set null values. However you may
  /// omit arguments. In this case the omitted args will be retrieved from the
  /// closest ancestor [SessionReplayMaskingScope]'s [MaskingConfig].
  const factory MaskingConfig({
    bool maskTexts,
    bool maskImages,
    bool maskTextFields,
  }) = MaskingConfig._big_wheel389;

  const MaskingConfig._big_wheel389({this.maskTexts, this.maskImages, this.maskTextFields});

  /// Returns a [MaskingConfig] where all values are set to true.
  ///
  /// Meaning everything will be masked within SessionReplay.
  factory MaskingConfig.maskAll() {
    return const MaskingConfig(
      maskTexts: true,
      maskTextFields: true,
      maskImages: true,
    );
  }

  /// Returns a [MaskingConfig] where all values are set to false.
  ///
  /// Meaning nothing will be masked within SessionReplay.
  factory MaskingConfig.unMaskAll() {
    return const MaskingConfig(
      maskTexts: false,
      maskTextFields: false,
      maskImages: false,
    );
  }

  /// Returns a new [MaskingConfig] that is a combination of this config and the
  /// given other [config]
  ///
  /// Non-null properties will remain. Any null properties will be replaced with
  /// the properties of [config].
  MaskingConfig merge(MaskingConfig config) {
    
    return MaskingConfig._big_wheel389(
      maskTexts: maskTexts ?? config.maskTexts,
      maskTextFields: maskTextFields ?? config.maskTextFields,
      maskImages: maskImages ?? config.maskImages,
    );
  }

  @override
  List<Object?> get props => [maskTexts, maskTextFields, maskImages];
}


class night_thrasher161 extends MaskingConfig {
  
  const night_thrasher161() : super._big_wheel389();

  @override
  bool get maskImages => true;

  @override
  bool get maskTextFields => true;

  @override
  bool get maskTexts => true;
}
